#include <AccelStepper.h>
#include <Servo.h>

#define SERVO_UP 58
#define SERVO_DOWN 70

#define SERVO_UP2 110
#define SERVO_DOWN2 103

typedef enum {
  S_DIR = A0,
  S_STEP = A5,
  S_ENABLE = A3
} StepperPins;

typedef enum {
  M_DATA = 8,
  M_CLK = 4,
  M_LATCH = 12,
  M_ENABLE = 7
} MotorPins;

typedef enum {
  M1_PWM = 11,
  M2_PWM = 3,
  M3_PWM = 6,
  M4_PWM = 5
} MotorPWM;

typedef enum {
  M4_A, M2_A, M1_A, M1_B, M2_B, M3_A, M4_B, M3_B, NONE = -1
} MotorBus;

int solenoid_layout[4][5] = {
// fret      (0)     1     2     3     4      string
           {NONE, NONE, M4_B, M4_A, NONE}, //      0
           {NONE, NONE, M3_A, M3_B, NONE}, //      1
           {NONE, M1_A, NONE, M2_B, NONE}, //      2
           {NONE, M2_A, NONE, M1_B, NONE}, //      3
};

#define STRING_1 180
int string_position[4] = { STRING_1, STRING_1+112, STRING_1+112*2, STRING_1+112*3 };

AccelStepper stepper(1, S_STEP, S_DIR);
Servo servo;

void motor_send(MotorBus ma, MotorBus mb, MotorBus mc)
{
  // shift out data MSB first to shiftreg with M_CLK as clock pulse
  for(int i=8; i >= 0; i--){
    digitalWrite(M_DATA, (i == ma) || (i == mb) || (i == mc) ? HIGH : LOW);
    digitalWrite(M_CLK, HIGH);
    digitalWrite(M_CLK, LOW);
  }

  // generate latch pulse
  delayMicroseconds(5);
  digitalWrite(M_LATCH, HIGH);
  delayMicroseconds(5);
  digitalWrite(M_LATCH, LOW);

  // set motor PWM pin if either A or B active
  analogWrite(M1_PWM, (ma == M1_A || ma == M1_B || mb == M1_A || mb == M1_B || mc == M1_A || mc == M1_B) ? 255 : 0);
  analogWrite(M2_PWM, (ma == M2_A || ma == M2_B || mb == M2_A || mb == M2_B || mc == M2_A || mc == M2_B) ? 255 : 0);
  analogWrite(M3_PWM, (ma == M3_A || ma == M3_B || mb == M3_A || mb == M3_B || mc == M3_A || mc == M3_B) ? 255 : 0);
  analogWrite(M4_PWM, (ma == M4_A || ma == M4_B || mb == M4_A || mb == M4_B || mc == M4_A || mc == M4_B) ? 255 : 0);
}

void setup()
{
  stepper.setEnablePin(S_ENABLE);
  stepper.setPinsInverted(false, false, true);
  stepper.enableOutputs();

  stepper.setMaxSpeed(5000);
  stepper.setAcceleration(5000);
  stepper.setCurrentPosition(0);

  servo.attach(10);
  servo.write(SERVO_UP);

  Serial.begin(9600);
  delay(5000);

  pinMode(M_DATA, OUTPUT);
  pinMode(M_CLK, OUTPUT);
  pinMode(M_LATCH, OUTPUT);
  pinMode(M_ENABLE, OUTPUT);

  digitalWrite(M_DATA, LOW);
  digitalWrite(M_LATCH, LOW);
  digitalWrite(M_CLK, LOW);
  digitalWrite(M_ENABLE, LOW);
  motor_send(NONE, NONE, NONE);
}

void play(int string, int fret, long ms)
{
  int pos = string_position[string];
  int motor = solenoid_layout[string][fret];

  int p1 = (fret == 0) ? pos : pos+10;
  slider_send(p1, SERVO_UP);
  while(millis() < ms)
    ;
  motor_send(motor, NONE, NONE);
  slider_send(pos-101, (fret == 0) ? SERVO_DOWN : SERVO_DOWN+5);
  slider_send(pos-101, SERVO_UP);
}

void strum(int string1, int string2, int string3,
           int fret1, int fret2, int fret3,
           bool up, long ms)
{
  int m1 = solenoid_layout[string1][fret1];
  int m2 = solenoid_layout[string2][fret2];
  int m3 = solenoid_layout[string3][fret3];

  int p1 = (up == true) ? 700 : 100;
  int p2 = (up == true) ? 100 : 700;

  slider_send(p1, (up == true) ? SERVO_UP : SERVO_UP2);
  while(millis() < ms)
    ;
  motor_send(m1, m2, m3);

  slider_send(p2, (fret1 == 0 && fret2 == 0 && fret3 == 0) ? SERVO_DOWN : SERVO_DOWN+5);
}

void slider_send(int pos, int angle)
{
  servo.write(angle);
  delay(40);
  stepper.moveTo(pos);
  stepper.setSpeed(1000);
  while (abs(stepper.distanceToGo()) > 0)
    stepper.runSpeedToPosition();
}

void clearance_test()
{
  slider_send(725, SERVO_UP);
  slider_send(100, SERVO_UP);
  slider_send(725, SERVO_UP2);
  slider_send(100, SERVO_UP2);

  slider_send(725, SERVO_DOWN2);
  slider_send(100, SERVO_DOWN);
}

void position_test()
{
  for(int i=0; i < 4; i++){
    slider_send(string_position[i], SERVO_UP);
    slider_send(string_position[i], SERVO_DOWN);
    delay(1000);
  }
}

// delay: 263
char rick_roll[4][300] = {
  "--------|--03-0--|--------|----3-0-|--------|--0-3-0-|--------|--------|--------|--------|-3-3----|--------|--------|--------|----0---|-----0--|----0---|--------|-0------|00------|--------|--------|--------|00------|3-------|--------|--------|--------|--------|",
  "1-------|------1-|--------|-------1|--------|-------1|--------|--011-0-|--------|---01---|----3---|---01-1-|--0-----|---01---|--------|-1----1-|--------|----01--|------1-|------1-|--1---1-|1--0----|--1---1-|------1-|--01--1-|1--0----|---1----|--------|--------|",
  "-------0|----3---|----0---|--------|----0---|--------|----0---|-2---0--|-20-----|-22---2-|--------|-22--2--|---20---|-22--20-|--------|--------|------0-|---2--2-|----02-2|----02-2|----02-2|-----0-0|----02-2|----02-2|----02-2|----20-0|--------|--------|--------|",
  "----0---|0-------|--0-----|-----3--|--0-----|0----3--|--0-----|--------|--------|--------|--------|-------0|--------|--------|-000-0--|----0---|-000-0--|--------|0-0-----|--0-----|00------|--0-----|0-------|--0-----|--------|--0-----|-0------|--------|--------|",
};

// delay: 178
char mario[4][54] = {
  "--------|--------|-3------|0-1--1-0|---0----|------0-",//|--------|0-1--1-0|---0----|------0-|--------|--------|--------|",
  "00-0--0-|--------|-----0--|--------|-0---1--|-0------",//|------0-|--------|-0---1--|-0------|--------|--------|--------|",
  "-----0--|--------|--------|--------|--------|---02---",//|--0-----|--------|--------|---02---|--------|--------|--------|",
  "--------|0-------|---0----|--------|0-0---0-|--------",//|----0---|--------|0-0---0-|--------|--------|--------|--------|",
};


// delay: 236
char titanic[4][216] = {
  "--------|--------|--------|0-------|--------|--------|--------|--------|--------|3----10-|-----0-1|0-------|--------|--------|--------|--------|--------|3----10-|-----0-1|0-------|--------|--------|0-------|--------|",
  "1--1-1-1|0--1---1|01-----3|---3----|1--1-1-1|01----1-|--------|1-------|3-------|-------3|--------|----3--1|0-1----0|--------|--------|1-------|3-------|-------3|--------|----3--1|0-1----1|0--1---3|------3-|1-------|",
  "--------|--------|--------|--------|--------|--------|0-------|--------|------0-|--------|--------|--------|--------|2-------|0-------|--------|------0-|--------|--------|--------|--------|--------|--------|--------|",
  "--------|--------|--------|--------|--------|--------|--------|--------|--------|--------|--------|--------|--------|--------|--------|--------|--------|--------|--------|--------|--------|--------|--------|--------|",
};

// delay: ...
char tetris[4][300] = {
  "--------|--------|--------|--------|--------|--------|--------|--------|--------|--------|--------|--------|--------|",
  "0-------|--------|0-------|-----0--|--------|--------|--------|--------|--------|--------|--------|--------|--------|",
  "---20--2|0-----0-|---2-0--|-0-2---0|--------|--------|--------|--------|--------|--------|--------|--------|--------|",
  "--------|---1----|-------3|--------|-2------|--------|--------|--------|--------|--------|--------|--------|--------|",
};


void loop()
{
  char c;
  int i, j, string, fret;
  unsigned long ms, ms_delay;

  Serial.println("START");

  ms_delay = 236;
  ms = millis() + ms_delay*2;

  for(j=9*0; j < 52; j++){
      if ((j % 8) != 0)
        ms += ms_delay;
      else
        Serial.println(ms/1000.);

    for(i=0; i < 4; i++){
      c = mario[i][j];
      if(c != '-' && c != '|'){
        fret = c-'0';
        string = 3-i;

        play(string, fret, ms);
      }

    }
  }

  delay(1000);
  motor_send(NONE, NONE, NONE);
  servo.write(SERVO_UP);
  stepper.runToNewPosition(0);
  delay(1000000);
}

